
# This file was generated by 'versioneer.py' (0.14) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

version_version = '0.44.0+0.g5bffb209e.dirty'
version_full = '5bffb209e853dc21a44a5bd801c93672404f1fe8.dirty'
def get_versions(default={}, verbose=False):
    return {'version': version_version, 'full': version_full}

